
#include "nanojit.h"


namespace nanojit
{


LirReader::LirReader(LirStorage* in) 
	: LirFilter(0),
	_in(in), 
	_it(in->getReverseBegin()),
	_finalIns(NULL)
{
	// The last instruction for a fragment shouldn't be a skip.
	// (Actually, if the last *inserted* instruction exactly fills up
	// a chunk, a new chunk will be created, and thus the last *written*
	// instruction will be a skip -- the one needed for the
	// cross-chunk link.  But the last *inserted* instruction is what
	// is recorded and used to initialise each LirReader, and that is
	// what is seen here, and therefore this assertion holds.)
	NanoAssert(!_it.getElement<LInsElement>()->data()->isop(LIR_skip));

	_finalIns = _in->getReverseBegin().getElement<LInsElement>()->data();
}

LirReader::~LirReader()
{

}

// Returns next instruction and advances to the prior instruction.
// Invariant: never returns a skip.
LIns* LirReader::read()
{
	while (_it.getElement<LInsElement>()->data()->isop(LIR_skip))
	{
		NanoAssert(_it != _in->getReverseEnd());
		_it.moveNext();
	}

	// Step back one instruction.  Use a table lookup rather than a switch
	// to avoid branch mispredictions.  LIR_start is given a special size
	// of zero so that we don't step back past the start of the block.
	// (Callers of this function should stop once they see a LIR_start.)
	LIns* ret = _it.getElement<LInsElement>()->data();

	_it.moveNext();

	return ret;
}

// Returns the instruction that read() will return on the next call.
// Invariant: never returns a skip.
LIns* LirReader::peek()
{
	// Check the invariant: _ins never points to a skip.
	NanoAssert(!_it.getElement<LInsElement>()->data()->isop(LIR_skip));
	return _it.getElement<LInsElement>()->data();
}

// LIR verbose specific
#ifdef NANOJIT_VERBOSE

const char* lirNames[] = {
#define OP___(op, repKind, retType, isCse) \
#op,
#include "LIRopcode.tbl"
#undef OP___
	NULL
};

const char* getLirName(LOpcode opcode)
{
	NanoAssert(size_t(opcode) < sizeof(lirNames) / sizeof(lirNames[0]));
	return lirNames[opcode];
}

#endif /* NANOJIT_VERBOSE */

// implementation
#ifdef NANOJIT_VERBOSE
void ReverseLister::finish()
{
	_logc->printf("\n");
	_logc->printf("=== BEGIN %s ===\n", _title);
	int j = 0;
	for (Seq<char*>* p = _strs.get(); p != NULL; p = p->tail)
		_logc->printf("  %02d: %s\n", j++, p->head);
	_logc->printf("=== END %s ===\n", _title);
	_logc->printf("\n");
}

LIns* ReverseLister::read() {
	// This check is necessary to avoid printing the LIR_start multiple
	// times due to lookahead in Assembler::gen().
	if (_prevIns && _prevIns->isop(LIR_start))
		return _prevIns;
	LIns* ins = in->read();
	InsBuf b;
	const char* str = _printer->formatIns(&b, ins);
	char* cpy = new (_alloc) char[strlen(str)+1];
	VMPI_strcpy(cpy, str);
	_strs.insert(cpy);
	_prevIns = ins;
	return ins;
}

	class RetiredEntry
	{
	public:
		Seq<LIns*>* live;
		LIns* i;
		RetiredEntry(): live(NULL), i(NULL) {}
	};

	//horrible in insert order map
	template <class K>
	class VectorMap
	{
	private:
		typedef PodVector< K > MapType;
		typedef typename PodVector< K >::Iter ItType;
		MapType map;
	public:
		VectorMap(Allocator* alloc) : map(alloc) {}
		bool    containsKey(K k)
		{
			for (size_t i = 0 ; i < map.size(); ++i)
				if (map[i] == k)
					return true;
			return false;
		}

		void put(K k)
		{
			map.push_back(k);
		}

		void remove(K k)
		{
			for (size_t i = 0 ; i < map.size(); ++i)
				if (map[i] == k)
					map[i] = 0;
		}

		class Iter
		{
		private:
			typedef VectorMap<K> MapType;
			const MapType& map;
			size_t current;
		public:
			Iter(MapType& map) : map(map), current(static_cast<size_t>(-1))
			{ }

			/** return true if more (k,v) remain to be visited */
			bool next() {
				do
				{
					current++;
				}
				while ((current < map.map.size()) && (map.map[current] == 0));
				return current < map.map.size();
			}

			/** return the current key */
			const K& key() const {
				return map.map[current];
			}
		};
	};

	class LiveTable
	{
		Allocator* alloc;
	public:
		VectorMap<LIns*> live;
		SeqBuilder<RetiredEntry*> retired;
		int retiredCount;
		int maxlive;
		LiveTable(Allocator* alloc)
			: alloc(alloc)
			, live(alloc)
			, retired(alloc)
			, retiredCount(0)
			, maxlive(0)
		{ }

		void add(LIns* ins, LIns* use) {
			NanoUnusedVariable(use);
			if (!ins->isImmAny() && !live.containsKey(ins)) {
				NanoAssert(size_t(ins->opcode()) < sizeof(lirNames) / sizeof(lirNames[0]));
				live.put(ins);
			}
		}

		void retire(LIns* i) {
			RetiredEntry *e = new (alloc) RetiredEntry();
			e->i = i;
			InsList livelist(alloc);
			VectorMap<LIns*>::Iter  iter(live);
			int live_count = 0;
			while (iter.next()) {
				LIns* ins = iter.key();
				if (!ins->isV()) {
					live_count++;
					livelist.insert(ins);
				}
			}
			e->live = livelist.get();
			if (live_count > maxlive)
				maxlive = live_count;

			live.remove(i);
			retired.insert(e);
			retiredCount++;
		}

		bool contains(LIns* i) {
			return live.containsKey(i);
		}
	};

	/*
	 * traverse the LIR buffer and discover which instructions are live
	 * by starting from instructions with side effects (stores, calls, branches)
	 * and marking instructions used by them.  Works bottom-up, in one pass.
	 * if showLiveRefs == true, also print the set of live expressions next to
	 * each instruction
	 */
	void live(LirFilter* in, Allocator* alloc, LInsPrinter *printer, LogControl *logc)
	{
		// traverse backwards to find live exprs and a few other stats.

		LiveTable live(alloc);
		int total = 0;
		//if (frag->lirbuf->state)
		//    live.add(frag->lirbuf->state, 0);
		for (LIns* ins = in->read(); !ins->isop(LIR_start); ins = in->read())
		{
			total++;

			// First handle instructions that are always live (ie. those that
			// don't require being marked as live), eg. those with
			// side-effects.  We ignore LIR_paramp.
			if (ins->isLive() && !ins->isLInsP())
			{
				live.add(ins, 0);
			}

			// now propagate liveness
			if (live.contains(ins))
			{
				live.retire(ins);

				switch (ins->opcode()) {
				case LIR_skip:
					NanoAssertMsg(0, "Shouldn't see LIR_skip");
					break;

				case LIR_start:
				case LIR_parami:
				CASE64(LIR_paramq:)
				case LIR_paramf:
				case LIR_paramd:
				case LIR_j:
				case LIR_immi:
				CASE64(LIR_immq:)
				case LIR_immd:
				case LIR_immf:
				case LIR_immf4:
				case LIR_comment:
				case LIR_entry:
					// No operands, do nothing.
					break;

				case LIR_label:
					break;

				case LIR_phireconcile:
					for (unsigned int iPhiValueIndex = 0; iPhiValueIndex < ins->phiCount(); ++iPhiValueIndex)
					{
						LIns* phiValue = ins->phiValues()[iPhiValueIndex];
						if (ins->phiData()->capturedStates[iPhiValueIndex].isLive)
						{
							live.add(phiValue, 0);
						}
					}

				case LIR_phicapture:
				{
					PhiState* phiStates = (PhiState*)alloc->alloc(sizeof(PhiState) * ins->phiCount());

					for (unsigned int iPhiValueIndex = 0; iPhiValueIndex < ins->phiCount(); ++iPhiValueIndex)
					{
						LIns* phiValue = ins->phiValues()[iPhiValueIndex];
						if (live.contains(phiValue))
						{
							phiStates[iPhiValueIndex].isLive = true;
						}
						else
						{
							phiStates[iPhiValueIndex].isLive = false;
						}
					}
					
					ins->phiData()->capturedStates = phiStates;

					break;
				}

				case LIR_regfence:
				case LIR_spillfence:
				case LIR_killfence:
				{
					break;
				}

				case LIR_ldi:
				case LIR_ldri:
				CASE64(LIR_ldq:)
				CASE64(LIR_ldrq:)
				case LIR_ldd:
				case LIR_ldf:
				case LIR_ldf4:
				case LIR_lduc2ui:
				case LIR_ldus2ui:
				case LIR_ldrus2ui:
				case LIR_ldc2i:
				case LIR_lds2i:
				case LIR_ldrs2i:
				case LIR_ldf2d:
				case LIR_livei:
				CASE64(LIR_liveq:)
				case LIR_lived:
				case LIR_livef:
				case LIR_livef4:
				case LIR_negi:
				case LIR_noti:
				case LIR_negd:
				case LIR_negf:
				case LIR_negf4:
				case LIR_absd:
				case LIR_absf:
				case LIR_absf4:
				case LIR_recipf:
				case LIR_recipf4:
				case LIR_rsqrtf:
				case LIR_rsqrtf4:
				case LIR_sqrtf:
				case LIR_sqrtf4:
				case LIR_sqrtd:
				CASE64(LIR_i2q:)
				CASE64(LIR_ui2uq:)
				case LIR_i2d:
				CASE64(LIR_q2d:)
				CASE64(LIR_d2q:)
				case LIR_ui2d:
				case LIR_i2f:
				case LIR_ui2f:
				case LIR_d2f:
				case LIR_f2d:
				case LIR_f2f4:
				case LIR_f4x:
				case LIR_f4y:
				case LIR_f4z:
				case LIR_f4w:
				case LIR_swzf4:
				CASE64(LIR_q2i:)
				case LIR_d2i:
				CASE64(LIR_dasq:)
				CASE64(LIR_qasd:)
				CASE86(LIR_modi:)
				case LIR_copyi:
				CASE64(LIR_copyq:)
				case LIR_copyf:
				case LIR_copyd:
				case LIR_copyf4:
					live.add(ins->oprnd1(), 0);
					break;

				case LIR_jt:
				case LIR_jf:
					live.add(ins->branchCondition(), 0);
					break;

					
				case LIR_exit:
				{
					ParameterValue* values      = ins->exitParamData();
					unsigned int    count           = ins->exitParamCount();
					for (unsigned int i = 0; i < count; ++i)
					{
						LIns* ins = values[i].value;
						if (ins != NULL)
							live.add(ins, 0);
					}
					break;
				}


				case LIR_sti:
				case LIR_stri:
				CASE64(LIR_stq:)
				CASE64(LIR_strq:)
				case LIR_std:
				case LIR_stf:
				case LIR_stf4:
				case LIR_sti2c:
				case LIR_sti2s:
				case LIR_stri2s:
				case LIR_std2f:
				case LIR_eqi:
				case LIR_nei:
				case LIR_lti:
				case LIR_gti:
				case LIR_lei:
				case LIR_gei:
				case LIR_ltui:
				case LIR_gtui:
				case LIR_leui:
				case LIR_geui:
				case LIR_eqd:
				case LIR_ltd:
				case LIR_gtd:
				case LIR_led:
				case LIR_ged:
				case LIR_eqf:
				case LIR_ltf:
				case LIR_gtf:
				case LIR_lef:
				case LIR_gef:
				case LIR_eqf4:
				CASE64(LIR_eqq:)
				CASE64(LIR_neq:)
				CASE64(LIR_ltq:)
				CASE64(LIR_gtq:)
				CASE64(LIR_leq:)
				CASE64(LIR_geq:)
				CASE64(LIR_ltuq:)
				CASE64(LIR_gtuq:)
				CASE64(LIR_leuq:)
				CASE64(LIR_geuq:)
				case LIR_lshi:
				case LIR_rshi:
				case LIR_rshui:
				case LIR_rori:
				case LIR_rcri:
				CASE64(LIR_lshq:)
				CASE64(LIR_rshq:)
				CASE64(LIR_rshuq:)
				CASE64(LIR_rorq:)
				CASE64(LIR_rcrq:)
				case LIR_addi:
				case LIR_addci:
				case LIR_subi:
				case LIR_subbi:
				case LIR_muli:
				CASE64(LIR_mulq:)
				CASE86(LIR_diviu:)
				CASE86(LIR_divis:)
				case LIR_addd:
				case LIR_subd:
				case LIR_muld:
				case LIR_divd:
				case LIR_addf:
				case LIR_subf:
				case LIR_mulf:
				case LIR_divf:
				case LIR_addf4:
				case LIR_subf4:
				case LIR_mulf4:
				case LIR_divf4:
				case LIR_dotf4:
				case LIR_dotf3:
				case LIR_dotf2:
				case LIR_minf4:
				case LIR_maxf4:
				case LIR_cmpgtf4:
				case LIR_cmpgef4:
				case LIR_cmpltf4:
				case LIR_cmplef4:
				case LIR_cmpeqf4:
				case LIR_cmpnef4:
				CASE64(LIR_addq:)
				CASE64(LIR_addcq:)
				CASE64(LIR_subq:)
				CASE64(LIR_subbq:)
				case LIR_andi:
				case LIR_andci:
				case LIR_ori:
				case LIR_xori:
				CASE64(LIR_andq:)
				CASE64(LIR_andcq:)
				CASE64(LIR_orq:)
				CASE64(LIR_xorq:)
				case LIR_file:
				case LIR_line:
					live.add(ins->oprnd1(), 0);
					live.add(ins->oprnd2(), 0);
					break;

				case LIR_cmovi:
				CASE64(LIR_cmovq:)
				case LIR_cmovd:
				case LIR_cmovf:
				case LIR_cmovf4:
					live.add(ins->oprnd1(), 0);
					live.add(ins->oprnd2(), 0);
					live.add(ins->oprnd3(), 0);
					break;

				case LIR_mullsi:
				case LIR_mullui:
				CASE64(LIR_mullsq:)
				CASE64(LIR_mulluq:)
					live.add(ins->oprnd1(), 0);
					live.add(ins->oprnd2(), 0);
					break;

				case LIR_mullshii:
				case LIR_mulluhii:
				CASE64(LIR_mullshiq:)
				CASE64(LIR_mulluhiq:)
					live.add(ins->oprnd1(), 0);
					break;

				case LIR_callv:
				case LIR_calli:
				CASE64(LIR_callq:)
				case LIR_calld:
				case LIR_callf:
				case LIR_callf4:
					for (int i = 0, argc = ins->callArgc(); i < argc; i++)
						live.add(ins->callArgs()[i].value, 0);
					break;

				case LIR_setflgz:
				case LIR_setflgs:
				case LIR_setflgc:
				case LIR_setflgv:
					live.add(ins->oprnd1(), 0);
					break;

				case LIR_getflgz:
				case LIR_getflgs:
				case LIR_getflgc:
				case LIR_getflgv:
				case LIR_getflgnz:
				case LIR_getflgns:
				case LIR_getflgnc:
				case LIR_getflgnv:
					live.add(ins->oprnd1(), 0);
					break;


				case LIR_cvt2irf:
				case LIR_cvt2itf:
				case LIR_cvt2ird:
				case LIR_cvt2itd:
					live.add(ins->oprnd1(), 0);
					break;

				case LIR_nop:
					break;

				default:
					NanoAssertMsgf(0, "unhandled opcode: %d", ins->opcode());
					break;
				}
			}
		}

		logc->printf("  Live instruction count %d, total %u, max pressure %d\n",
					 live.retiredCount, total, live.maxlive);
		logc->printf("  Showing LIR instructions with live-after variables\n");
		logc->printf("\n");

		// print live exprs, going forwards
		bool newblock = true;
		for (Seq<RetiredEntry*>* p = live.retired.get(); p != NULL; p = p->tail) {
			RetiredEntry* e = p->head;
			InsBuf ib;
			RefBuf rb;
			static char livebuf[4096];
			char *s=livebuf;
			*s = 0;
			if (!newblock && e->i->isLInsLabel()) {
				logc->printf("\n");
			}
			newblock = false;
			for (Seq<LIns*>* p = e->live; p != NULL; p = p->tail) {
				VMPI_strcpy(s, printer->formatRef(&rb, p->head));
				s += VMPI_strlen(s);
				*s++ = ' '; *s = 0;
				NanoAssert(s < livebuf+sizeof(livebuf));
			}
			/* If the LIR insn is pretty short, print it and its
			   live-after set on the same line.  If not, put
			   live-after set on a new line, suitably indented. */
			const char* insn_text = printer->formatIns(&ib, e->i);
			if (VMPI_strlen(insn_text) >= 30-2) {
				logc->printf("  %-30s\n  %-30s %s\n", insn_text, "", livebuf);
			} else {
				logc->printf("  %-30s %s\n", insn_text, livebuf);
			}

			if (e->i->isLInsBranch()) {
				logc->printf("\n");
				newblock = true;
			}
		}
	}

#endif


} // namespace nanojit

